from typing import Dict, List, Optional, Tuple
from decimal import Decimal
from enum import Enum

from .utils import Executable
from .pipelines import StraightPipeline, Pipeline, Context


__all__ = 'Filter', 'FilterLayer',


class FilterLayer(int, Enum):
    PRECEDE = 100
    AVERAGE = 500
    FINAL = 900


class Filter:
    Layer = FilterLayer
    INCREMENT: Decimal = Decimal('0.0000000000001')

    default_priority: int = 10
    pipeline_class: Pipeline = StraightPipeline
    pipeline: Pipeline
    version: Tuple[int, int]
    registry: Dict[Executable, Tuple[int, int]]

    def __init__(
        self,
        initial: List[Executable] = [],
        pipeline_class: Optional[Pipeline] = None,
        default_priority: Optional[int] = None,
    ):
        self.pipeline_class = pipeline_class or self.pipeline_class
        self.default_priority = default_priority or self.default_priority
        self.version = [0, self.INCREMENT]
        self.registry = {}

        self.add_multiple(initial)

    def is_fresh(self) -> bool:
        cached, current = self.version

        return cached == current

    def resolve_pipeline(self):
        if self.is_fresh():
            return self.pipeline

        self.pipeline = self.pipeline_class(
            sorted(self.registry, key=self.order_key)
        )
        self.version[0] = self.version[1]

        return self.pipeline

    def order_key(self, entry: Executable) -> Tuple:
        return self.registry[entry]

    def add(
        self,
        fn: Executable,
        priority: Optional[int] = None,
        layer: Optional[int] = None,
    ):
        if fn in self.registry:
            raise ValueError(
                f'Handler "{fn}" already registered. Remove it to add again.'
            )

        self.version[1] += self.INCREMENT
        priority = Decimal(priority) if priority is not None else self.default_priority

        self.registry[fn] = (
            layer if layer is not None else self.Layer.AVERAGE,
            priority + self.version[1],
        )

    def add_multiple(
        self,
        fns: List[Executable],
        priority: Optional[int] = None,
        layer: Optional[int] = None,
    ):
        for item in fns:
            self.add(item, priority=priority, layer=layer)

    def remove(self, fn: Executable):
        if fn not in self.registry:
            raise ValueError(f'Handler "{fn}" not registered.')

        self.registry.pop(fn)

    def __call__(self, context: Context) -> Context:
        return self.resolve_pipeline()(context)



from typing import Any, Callable, List, Optional, Tuple

from .exceptions import StopFlow
from .utils import normalize_executable, Executable


__all__ = 'Pipeline', 'StraightPipeline',

Context = dict
Step = Callable[..., Any]


class Pipeline:
    START = object()
    FINISH = object()

    def __init__(self, pipeline):
        self.pipeline = pipeline

    def next(self, current, context: Context) -> Tuple[Any, Optional[Step]]:
        raise NotImplementedError()

    def prepare_context(self, context: Context) -> Context:
        return context

    def handle_result(self, step: Step, result: Any, context: Context) -> Any:
        if result is None:
            return context

        if isinstance(result, dict):
            context.update(result)

        return context

    def execute_step(self, step: Step, context: Context) -> Any:
        return step(context)

    def __call__(self, context: Context) -> Context:
        context = self.prepare_context(context)
        current = self.START

        while current != self.FINISH:
            try:
                current = self.next(current, context)
                result = self.execute_step(current, context)
            except StopFlow as e:
                result = e.result
                current = self.FINISH

            context = self.handle_result(current, result, context)

        return context


StraightPipelineSteps = List[Executable]
StraightStep = Tuple[int, Optional[Step]]


class StraightPipeline(Pipeline):
    pipeline: StraightPipelineSteps

    def execute_step(self, step: StraightStep, context: Context) -> Any:
        return step[1](context)

    def next(self, current, context: Context) -> StraightStep:
        index = -1

        if current is not self.START:
            index, _ = current

        if index + 1 >= len(self.pipeline):
            raise StopFlow(None)

        index += 1

        return index, normalize_executable(self.pipeline[index])



from functools import lru_cache
from importlib import import_module
from typing import Any, Callable, Union


Executable = Union[str, Callable[..., Any]]


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
        ) from err


cached_import_string = lru_cache(maxsize=None)(import_string)


def normalize_executable(executable: Executable) -> Executable:
    if isinstance(executable, str):
        return cached_import_string(executable)

    return executable



class PipelineError(Exception):
    """
    Base class for exceptions in this module.
    """
    pass


class StopFlow(PipelineError):
    """
    Raised when a flow should be stopped immediately.
    """

    def __init__(self, result, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.result = result
